/* This file is part of smap.
   Copyright (C) 2006, 2007, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "smap.h"

static int
copy_part(const char *cstr, const char *p, char **pbuf)
{
	size_t len = p - cstr;
	char *buf = malloc(len + 1);
	if (!buf) {
		free(buf);
		errno = ENOMEM;
		return -1;
	}
	memcpy(buf, cstr, len);
	buf[len] = 0;
	*pbuf = buf;
	return 0;
}
	

static int
split_url(const char *proto, const char *cstr, char **pport, char **ppath)
{
	const char *p = strchr(cstr, ':');

	if (!p) {
		*pport = NULL;
		*ppath = strdup(cstr);
		return *ppath == NULL;
	} else if (copy_part(cstr, p, ppath))
		return -1;
	else
		cstr = p + 1;
	*pport = strdup(cstr);
	return *pport == NULL;
}

int
split_connection(const char *cstr, 
		 char **pproto, char **pport, char **ppath)
{
	const char *p;
	
	p = strchr(cstr, ':');
	if (!p) {
		*pproto = NULL;
		if (cstr[0] == '/') {
			if ((*pproto = strdup("unix")) == NULL)
				return -1;
			if ((*ppath = strdup(cstr)) == NULL) {
				free(*pproto);
				errno = ENOMEM;
			}
			*pport = 0;
		} else {
			errno = EINVAL;
			return -1;
		}
	} else {
		char *proto;
		if (copy_part(cstr, p, &proto))
			return -1;
		cstr = p + 1;
		if (cstr[0] == '/' && cstr[1] == '/') {
			int rc = split_url(proto, cstr + 2, pport, ppath);
			if (rc)
				free(proto);
			else
				*pproto = proto;
			return rc;
		}
		*pproto = proto;
	}
	return 0;
}

int
parse_url0(const char *cstr,
	   char *proto, char *port, char *path,
	   struct sockaddr **psa, socklen_t *plen)
{
	all_addr_t addr;
	socklen_t socklen;
	struct sockaddr *sa;

	if (!proto
	    || strcmp(proto, "unix") == 0 || strcmp(proto, "local") == 0) {
		if (port) {
			smap_error("invalid connection type: %s; "
				   "port is meaningless for UNIX sockets",
				   cstr);
			return -1;
		}
		
		if (strlen(path) > sizeof addr.s_un.sun_path) {
			errno = EINVAL;
			smap_error("%s: UNIX socket name too long", path);
			return -1;
		}
		
		addr.sa.sa_family = PF_UNIX;
		socklen = sizeof(addr.s_un);
		strcpy(addr.s_un.sun_path, path);
	} else if (strcmp(proto, "inet") == 0) {
		short pnum;
		long num;
		char *p;
		
		addr.sa.sa_family = PF_INET;
		socklen = sizeof(addr.s_in);

		if (!port) {
			smap_error("invalid connection type: %s; "
				   "missing port number",
				   cstr);
			return -1;
		}

		num = pnum = strtol(port, &p, 0);
		if (*p == 0) {
			if (num != pnum) {
				smap_error("invalid connection type: "
					   "%s; bad port number", cstr);
				return -1;
			}
			pnum = htons(pnum);
		} else {
			struct servent *sp = getservbyname(port, "tcp");
			if (!sp) {
				smap_error("invalid connection type: "
					   "%s; unknown port name",
					   cstr);
				return -1;
			}
			pnum = sp->s_port;
		}
		
		if (!path)
			addr.s_in.sin_addr.s_addr = INADDR_ANY;
		else {
			struct hostent *hp = gethostbyname(path);
			if (!hp) {
				smap_error("unknown host name %s", path);
				return -1;
			}
			addr.sa.sa_family = hp->h_addrtype;
			switch (hp->h_addrtype) {
			case AF_INET:
				memmove(&addr.s_in.sin_addr, hp->h_addr, 4);
				addr.s_in.sin_port = pnum;
				break;

			default:
				smap_error("invalid connection type: "
					   "%s; unsupported address family",
					   cstr);
				return -1;
			}
		}
	} else {
		smap_error("unsupported protocol: %s", proto);
		return -1;
	}
	sa = emalloc(socklen);
	memcpy(sa, &addr, socklen);
	*psa = sa;
	*plen = socklen;
	return 0;
}

int
parse_url(const char *cstr, struct sockaddr **psa, socklen_t *plen)
{
	int rc;
	char *proto, *port, *path;
	
	if (split_connection(cstr, &proto, &port, &path)) {
		smap_error("cannot parse url %s", cstr);
		return 1;
	}
	rc = parse_url0(cstr, proto, port, path, psa, plen);
	free(proto);
	free(port);
	free(path);
	return rc;
}

